program sphericaloverdensity
  !---------------------------------------------------------------------
  ! Ce programme calcule la fonction de masse avec
  ! l'algorithme SPHERICAL OVERDENSITY.
  ! Ce programme doit lire en input les fichiers suivants:
  !          - un fichier resultat RAMSES:    part_00600.out
  ! Il genere en output les fichiers suivants:
  !          - un fichier fonction de masse:  mf_sod_00600.dat
  !          - un fichier structure:          st_sod_00600.dat
  ! E. Audit, R. Teyssier
  ! Meudon, le 30/05/96.
  !---------------------------------------------------------------------
  ! Version F90 par R. Teyssier le 28/02/00   
  ! f90 sod.f90 -o sod
  ! Memory: 44*npart in bytes
  !         npart=256**3   707 Mo
  !         npart=512**3   5.6 Go
  !---------------------------------------------------------------------
  implicit none
  integer::ndim,npart,ngrid,n,i,icpu,ipos
  integer::ny,nz,ncpu,npart_new
  integer::ncpu2,npart2,ndim2,option
  real(kind=8)::r,scale,Mmin,mtot,rhomoyen
  real::Lbox,hubble,omega0,omegaL,aexp,mass_in_sol
  real,dimension(:,:),allocatable::x
  real,dimension(:,:),allocatable::pos,vel
  real,dimension(:),allocatable::mp,tmp
  integer,dimension(:),allocatable::isort
  real(kind=8),dimension(:),allocatable::xsort

  real(kind=8)::seuil=600.
  integer::nmin=100,nx=0,ncut=4

  character*5::nchar,ncharcpu
  character*50::nomfich,repository
  logical::ok,verbose,megaverbose

  call read_params
  ncpu=-1
  option=999
  verbose=.false.
  megaverbose=.false.
  !-----------------------------------------------
  ! Lecture du fichier particules au format GADGET
  !-----------------------------------------------
  call read_gadget(option,repository,ncpu,verbose,megaverbose, &
 &                 Lbox,hubble,omega0,omegaL,aexp,mass_in_sol, &
 &                 npart)

  write(*,*)'Found ',npart,' particles'
  write(*,*)'Reading positions and masses...'

  deallocate(vel)
  allocate(x(1:npart,1:3))

  do i=1,npart
     x(i,1:3)=pos(1:3,i)/Lbox
  enddo
  deallocate(pos)

  allocate(mp(1:npart))
  do i=1,npart
     mp(i)=1./dble(npart)
  enddo
  mtot=1.
  if(nx==0)then
     nx=int(dble(npart)**(1./3.))
  endif

  allocate(xsort(1:npart))
  allocate(isort(1:npart))
  ny=nx
  nz=nx
  rhomoyen=mtot
  write(*,*)'Working mesh=',nx,ny,nz
  write(*,*)'Found rho_bar=',rhomoyen,' -> rescaling to 1.0'
  mp=mp/rhomoyen

  scale=nx
  x=x*scale
  mp=mp*scale**3
  Mmin=dble(nmin)/dble(npart)*scale**3

  ngrid=nx*ny*nz

  !------------------------------------------
  ! Tri des particules selon leur densite NGP
  !------------------------------------------
  write(*,*) 'Sorting particles by NGP density'
  call tri_ngp(x,mp,xsort,isort,npart,nx,ny,nz,npart_new)

  !---------------------------------------------------------
  ! Calcul de la fonction de masse par SPHERICAL OVERDENSITY
  !---------------------------------------------------------
  write(*,*) 'Computing halo mass function'
  write(*,*) 'using a spherical overdensity delta_bar=',seuil
  call spherover(x,mp,xsort,isort,repository,npart,nx,ny,nz,npart_new,ncut,seuil,Mmin,scale)
  write(*,*) 'done'

  stop

contains

  subroutine read_gadget(option,filein,nmpi,verbose,megaverbose, &
 &                     Lbox,hubble,omega0,omegaL,aexp,mass_in_sol, &
 &                     npart)
  !=======================================================================
  ! This routine reads positions and velocities in a (multiple) GADGET
  ! file.
  !
  ! INPUTS :
  ! --------
  ! option (integer)       : 1 to read only header of the files, 
  !                          other number to read the full data
  ! filein (string)        : name of the input file (see below)
  ! nmpi   (integer)       : number of outputs per snapshot. The routine 
  !          works only for nmpi <= 1000.
  !          - if nmpi=-1, then there is only one input file with name filein
  !          - if 2 <= nmpi <= 1000, then there are nmpi input files with 
  !             extension impi where 0 <= impi <= nmpi-1. 
  !          - Exemple : filein='toto.', nmpi=3. The input files read are
  !            toto.0, toto.1 and toto.2
  !          - Note : nmpi=1 is not allowed 
  ! verbose (logical)      : .true. for verbose mode, .false. otherwise
  ! megaverbose(logical)   : .true. for detailed verbose mode, 
  !                          .false. otherwise
  !
  ! OUTPUTS :
  ! ---------
  ! Lbox   (real 4)        : size of the simulation cube in Mpc
  ! hubble (real 4)        : H0/100 where H0 is the present time Hubble 
  !                          constant in km/s/Mpc
  ! Omega0 (real 4)        : density parameter
  ! OmegaL (real 4)        : cosmological constant
  ! aexp   (real 4)        : expansion factor normalized to unity at present time
  ! mass_in_sol (real 4)   : mass of each particle in solar masses
  ! npart (integer)        : number of particles in the simulation
  ! ((xp(i,j),i=1,3),j=1,npart) (real 4) : particles coordinates in Mpc
  !                         0 <= xp(1,:) < Lbox
  !                         0 <= xp(2,:) < Lbox
  !                         0 <= xp(3,:) < Lbox
  ! ((vp(i,j),i=1,3),j=1,npart) (real 4) : particles velocities in km/s
  !
  ! NOTE : 
  ! ------
  ! xp and vp are global module variables.
  ! These allocatable arrays must be unallocated prior to the call of this
  ! routine. 
  !=======================================================================
    implicit none
    character(len=*) :: filein
    integer :: npart,nmpi
    logical :: verbose,megaverbose
    real(kind=4) :: aexp,omega0,omegaL,hubble,Lbox,mass_in_sol
    integer :: option
  !=======================================================================

    character(len=1024) :: filempi
    character(len=6) :: string
    integer :: id,jd,kd
    integer :: npartold,npartp,impi,i,j
    integer :: npart_in(0:5), nall_in(0:5)
    real(kind=8) :: massarr_in(0:5),a_in,redshift_in
    real(kind=8) :: omega0_in,omegaL_in,hubble_in
    integer :: unused_in(64-6-12-2-2-1-1-6-1-1-2-2-2-2)
    integer :: flag_sfr_in,flag_feedback_in,flag_cooling_in,numfiles_in
    real(kind=8) :: xLbox_in
    real(kind=8) :: facco,mass_in_kg
    integer, allocatable, dimension(:) :: indx
    real(kind=4), allocatable, dimension(:) :: tmp
    
    integer, parameter :: lin=10
    integer, parameter :: optioncheck=1
    
    ! Physical constants (units : m s kg) ->
    real(kind=8), parameter :: critical_density= 1.8788d-26
    real(kind=8), parameter :: mega_parsec=3.0857d22
    real(kind=8), parameter :: solar_mass=1.989d30
    ! <-

    if (verbose) then
       if (option==optioncheck) then
          write(*,*) 'Checking '//trim(filein)
       else
          write(*,*) 'Reading '//trim(filein)
       endif
    endif
        
    if (nmpi==1) then
       write(*,*) 'ERROR in read_gadget'
       write(*,*) 'nmpi=1 does not make any sense.'
       STOP
    elseif (nmpi > 1000) then
       write(*,*) 'ERROR in read_gadget'
       write(*,*) 'nmpi > 1000 not yet implemented.'
       STOP       
    endif

    npart=0
    do impi=0,abs(nmpi)-1
       if (nmpi==-1) then
          filempi=trim(filein) 
       else
          call convtoasc(impi,string)
          if (impi < 10) then
             filempi=trim(filein)//string(6:6)
          elseif (impi < 100) then
             filempi=trim(filein)//string(5:6)
          else
             filempi=trim(filein)//string(4:6)
          endif
       endif
       if (megaverbose)  write(*,*) 'Read header of '//trim(filempi)
       open(unit=lin,file=filempi,form='unformatted',status='old')
       read(lin) (npart_in(i),i=0,5), (massarr_in(i),i=0,5), a_in,  &
 &             redshift_in, flag_sfr_in, flag_feedback_in,  &
 &             (nall_in(i),i=0,5), flag_cooling_in, numfiles_in, &
 &             xLbox_in,omega0_in,omegaL_in,hubble_in, &
 &             unused_in
       close(lin)
       npartp=sum(npart_in)
       npart=npart+npartp
       if (megaverbose) then          
          write(*,*) 'icpu,npartp,aexp=',impi,npartp,a_in
       endif
       if (numfiles_in > 1.and.abs(nmpi)==1) then
          write(*,*) 'ERROR in read_part :'
          write(*,*) 'Only pure dark matter simulations for GADGET'
          write(*,*) 'numfiles_in=',numfiles_in
          STOP
       endif
    enddo

    aexp=a_in
    omega0=omega0_in
    omegaL=omegaL_in
    hubble=hubble_in
    facco=1.d0/(1000.d0*hubble_in)
    Lbox=xLbox_in*facco
    ! mass of each particle in kg
    mass_in_kg=((xLbox_in*facco)**3/dble(npart))*mega_parsec**3 &
 &             *omega0_in*critical_density*hubble_in**2
    
    ! mass of each particle in solar mass
    mass_in_sol=real(mass_in_kg/solar_mass)
    
    if (megaverbose) then
       write(*,*) 'npart,aexp,omega0,omegaL,hubble,Lbox='
       write(*,*) npart,aexp,omega0,omegaL,hubble,Lbox
    endif

    if (verbose) then
       write(*,*) '------------------'
       write(*,*) 'SIMULATION INFOS :'
       write(*,*) '------------------'
       write(*,*)
       write(*,*) 'Lbox (Mpc) =',Lbox
       write(*,*) 'aexp       =',aexp
       write(*,*) 'H0/100     =',hubble
       write(*,*) 'Omega0     =',Omega0
       write(*,*) 'OmegaLambda=',OmegaL
       write(*,*) 'num or part=',npart
       write(*,*) 'Mass of each part (solar mass)=',mass_in_sol
       write(*,*)
    endif
    if (option==optioncheck) return
    
    if (nmpi /= -1) allocate(indx(npart),tmp(npart))
    
    allocate(pos(3,npart),vel(3,npart))
    npart=0
    npartold=1
    do impi=0,abs(nmpi)-1
       if (nmpi==-1) then
          filempi=trim(filein) 
       else
          call convtoasc(impi,string)
          if (impi < 10) then
             filempi=trim(filein)//string(6:6)
          elseif (impi < 100) then
             filempi=trim(filein)//string(5:6)
          else
             filempi=trim(filein)//string(4:6)
          endif
       endif
       open(unit=lin,file=filempi,form='unformatted',status='old')
       read(lin) (npart_in(i),i=0,5), (massarr_in(i),i=0,5), a_in,  &
 &             redshift_in, flag_sfr_in, flag_feedback_in,  &
 &             (nall_in(i),i=0,5), flag_cooling_in, numfiles_in, &
 &             xLbox_in,omega0_in,omegaL_in,hubble_in, &
 &             (unused_in(i),i=1,64-6-12-4-2*4-6*4-2*4-4*8)
       npart=npart+sum(npart_in)
       if (megaverbose)  write(*,*) 'Read '//trim(filempi)
       if (megaverbose) write(*,*) 'Read particle coordinates...'
       read(lin) ((pos(i,j),i=1,3),j=npartold,npart)
       if (megaverbose) write(*,*) 'Read particle velocities...'
       read(lin) ((vel(i,j),i=1,3),j=npartold,npart)
       if (nmpi/=-1) then
          if (megaverbose) write(*,*) 'Read particle ids...'
          read(lin) (indx(j),j=npartold,npart)
       endif
       close(lin)     
       npartold=npart+1
    enddo
  

    ! Reshuffle particles according to their real order
    if (nmpi /= -1) then
       if (megaverbose) write(*,*) 'Reorder the particles (this might take a while)...'
       do j=1,3
          tmp(1:npart)=pos(j,1:npart)
          do i=1,npart
             pos(j,indx(i))=tmp(i)
          enddo
          tmp(1:npart)=vel(j,1:npart)
          do i=1,npart
             vel(j,indx(i))=tmp(i)
          enddo
       enddo
       deallocate(tmp,indx)
    endif

    ! Positions are in Mpc, in a box of size xlong, ylong, zlong
    ! Origin of coordinates is at the left corner of the box
    if (megaverbose) write(*,*) 'renormalize positions...'
    pos(1,1:npart)=pos(1,1:npart)*facco
    pos(2,1:npart)=pos(2,1:npart)*facco
    pos(3,1:npart)=pos(3,1:npart)*facco
    if (megaverbose) then
       write(*,*) 'X(1:5)='
       write(*,*) pos(1,1:5)
       write(*,*) 'Y(1:5)='
       write(*,*) pos(2,1:5)
       write(*,*) 'Z(1:5)='
       write(*,*) pos(3,1:5)
    endif
    
    ! Convert velocities in physical peculiar velocities in km/s
    if (megaverbose) write(*,*) 'renormalize velocities...'
    vel(1,1:npart)=vel(1,1:npart)*sqrt(a_in)
    vel(2,1:npart)=vel(2,1:npart)*sqrt(a_in)
    vel(3,1:npart)=vel(3,1:npart)*sqrt(a_in)
    if (megaverbose) then
       write(*,*) 'VX(1:5)='
       write(*,*) vel(1,1:5)
       write(*,*) 'VY(1:5)='
       write(*,*) vel(2,1:5)
       write(*,*) 'VZ(1:5)='
       write(*,*) vel(3,1:5)
    endif
    
    if (verbose) write(*,*) 'The input file has been read successfully'

  end subroutine read_gadget
  !=======================================================================
  subroutine convtoasc(number,sstring)
  !=======================================================================
  ! To convert an integer smaller than 999999 to a 6 characters string
  !=======================================================================
    implicit none
    integer :: number, istring, num, nums10, i
    character(len=6) :: sstring
    character(len=10), parameter :: nstring='0123456789'
    
    num=1000000
    nums10=num/10
    do i=1,6
       istring=1+mod(number,num)/nums10
       sstring(i:i)=nstring(istring:istring)
       num=num/10
       nums10=nums10/10
    enddo
  end subroutine convtoasc


  subroutine read_params

      implicit none

      integer       :: i,n
      integer       :: iargc
      character(len=4)   :: opt
      character(len=128) :: arg
      LOGICAL       :: bad, ok
      
      n = iargc()
      if (n < 2) then
         print *, 'usage: sod  [-inp input_file]'
         print *, '            [-nx  nx_grid]  (optional)'
         print *, '            [-min np_min]   (optional)'
         print *, '            [-sel ncell]    (optional)'
         print *, '            [-del del_bar]  (optional)'
         print *, '            [-hlp]          (optional)'
         stop
      end if

      do i = 1,n,2
         call getarg(i,opt)
         if (opt == '-hlp') then
            print 1, repository,nx,seuil,mmin
            stop
         end if
         if (i == n) then
            print '("option ",a2," has no argument")', opt
            stop 2
         end if
         call getarg(i+1,arg)
         select case (opt)
         case ('-inp')
            repository = trim(arg)
         case ('-nx')
            read (arg,*) nx
         case('-min')
            read (arg,*) nmin
         case('-sel')
            read (arg,*) ncut
         case ('-del') 
            read (arg,*) seuil
         case default
            print '("unknown option ",a2," ignored")', opt
         end select
      end do

1     format(/,&
           & " -inp [] ",A,/,&
           & " -nx  [] ",I6,/,&
           & " -min [] ",I6,/,&
           & " -sel [] ",I6,/,&
           & " -del [] ",e10.2,/)

      return

    end subroutine read_params

end program sphericaloverdensity

!------------------------------------------------------------
! TRI DES PARTICULES SELON LA DENSITE NGP SUR UNE GRILLE FINE
!------------------------------------------------------------
subroutine tri_ngp(x,mp,pdens,isort,npart,nx,ny,nz,npart_new)
  implicit none
  integer::npart,nx,ny,nz,npart_new
  integer,dimension(1:npart)::isort
  real,dimension(1:npart,1:3)::x
  real,dimension(1:npart)::mp
  real(kind=8),dimension(1:npart)::pdens

  integer(kind=8),dimension(1:npart)::indx
  integer::ngxngy,ngx,ngy,ngz
  integer(kind=8)::nmax,i1,i2,i3,ind
  integer::i,j,ifail
  integer::imin,imax,ntot
  real(kind=8)::amass
  
  ngx=4*nx
  ngy=4*ny
  ngz=4*nz

  ngxngy=ngx*ngy
   
  ! Remplissage du tableau des indices
  !-----------------------------------
  do i = 1,npart
     i1 = int(x(i,1)*dble(ngx)/dble(nx))
     i2 = int(x(i,2)*dble(ngy)/dble(ny))
     i3 = int(x(i,3)*dble(ngz)/dble(nz))
     indx (i) = 1+i1+i2*ngx+i3*ngxngy
  end do
  
  ! Tri des particules selon les indices
  !-------------------------------------
  ifail=0
  pdens=indx
  call quick_sort(pdens,isort,npart)
  indx=pdens

  ! Calcul de l'inverse de la densite NGP pour chaque particule
  !------------------------------------------------------------
  imin = 1
  do while(imin .le. npart)
     ind  = indx(imin)
     i    = imin
     do while(indx(i).eq.ind.and.i.lt.npart)
        i = i + 1
     end do
     if(i.eq.npart)i=npart+1
     imax = i - 1
     ntot = i - imin
     amass= 0.0
     do i = imin ,imax
       amass=amass+mp(isort(i))
     end do
     do i = imin ,imax
        pdens(isort(i)) = 1.0d0/amass
     end do
     imin = imax + 1
  end do
  print *, 1./minval(pdens)
  print *, 1./maxval(pdens)

  ! Tri des particules selon la densite
  !------------------------------------
  call quick_sort(pdens,isort,npart)

  i=1
  do while (pdens(i) < 0.5)
     i=i+1
  end do
  npart_new=i
  write(*,*)'npart_active=',npart_new
  return

end subroutine tri_ngp

!---------------------------------------------------------------------------
subroutine spherover(x,mp,xsort,isort,repository,npart,nx,ny,nz,npart_new,ncut,seuil,Mmin,scale)
!---------------------------------------------------------------------------
  implicit none

  character*50::repository
  integer::npart,nx,ny,nz,npart_new
  real(kind=8)::seuil,Mmin,scale
  integer::ncut          ! Amplitude de selection des particules
  real,dimension(1:npart,1:3)::x
  real,dimension(1:npart)::mp
  real(kind=8),dimension(1:npart)::xsort
  integer,dimension(1:npart)::isort
  
  integer::np
  real(kind=8)::twopi=6.283185307179586
  real(kind=8)::dtmax=0.01 ! dr/rayon pour la convergence du barycentre
  integer::Nsel=4000000    ! Nombre maximum de particules selectionnees

!  integer,dimension(1:npart)::structure
  integer,dimension(1:npart)::indx_ngp,isort_ngp
  integer,dimension(:),allocatable::first_ngp,num_ngp

  integer,dimension(:),allocatable::indx_sel,isort_sel,irank_sel
  real,dimension(:),allocatable::m_sel
  real,dimension(:,:),allocatable::x_sel
  real(kind=8),dimension(:),allocatable::distance

  integer::i,j,k,npart_sel,ipart  
  integer::is,iamas,masse,ip,ifail
  integer::ic,jc,kc,ii,jj,kk,ind
  integer::i1,i2,i3,imin,imax,ntot
  integer::niter,itest
  real(kind=8)::xc,yc,zc,xx,yy,zz,xc0,yc0,zc0,dc0
  real(kind=8)::xb,yb,zb,dt
  real(kind=8)::taillex,tailley,taillez,amass
  real(kind=8)::dx,dy,dz,d,d2,rayon,rayon2,volume,overdens
  real(kind=8)::dmin,x_tmp,y_tmp,z_tmp,dis_tmp,qqq
  real(kind=8)::Mtotsel

  character*80::nomf

!  structure = 0
  taillex   = dble(nx)
  tailley   = dble(ny)
  taillez   = dble(nz)
  nomf      = 'sod_'//TRIM(repository)//'.dat'

  allocate(num_ngp  (1:nx*ny*nz))
  allocate(first_ngp(1:nx*ny*nz))

  allocate(indx_sel (Nsel))
  allocate(isort_sel(Nsel))
  allocate(distance (Nsel))
  allocate(x_sel    (Nsel,1:3))
  allocate(m_sel    (Nsel))

  ! Calcul du pointeur cellules vers particules
  !--------------------------------------------
  write(*,*)'Building particle linked list...'
  do i  = 1,npart
     i1 = int(x(i,1))
     i2 = int(x(i,2))
     i3 = int(x(i,3))
     indx_ngp(i) = 1+i1+i2*nx+i3*nx*ny
  end do

  xsort=indx_ngp
  call quick_sort(xsort,isort_ngp,npart)
  indx_ngp=xsort

  first_ngp(:)=0
  num_ngp(:)=0
  imin = 1
  do while(imin.le.npart)
     ind  = indx_ngp(imin)
     i    = imin
     first_ngp(ind) = imin
     do while(indx_ngp(i).eq.ind.and.i.lt.npart)
        i = i + 1
     end do
     if(i.eq.npart)i=npart+1
     imax = i - 1
     ntot = i - imin
     num_ngp(ind) = ntot
     imin = imax + 1
  end do

  write(*,*) '  halo#    npart  mass       x          y          z         it  radius     del'
  open(10,file=nomf,form='formatted')
  iamas = 0
  xsort = 1.0
  do ipart = 1,npart_new

     ! On choisit la particule la plus dense de la liste
     !-------------------------------------------------
     if(xsort(isort(ipart)).gt.0.)then

        ! On initialise le barycentre de l'amas avec 
        ! la position de la particule courante
        xc = x(isort(ipart),1)
        yc = x(isort(ipart),2)
        zc = x(isort(ipart),3)

        xc0=xc; yc0=yc; zc0=zc

        ! Preselection des particules les plus proches du barycentre
        !-----------------------------------------------------------
        is = 0
        ic = int(xc)
        jc = int(yc)
        kc = int(zc)
        mtotsel=0.
        do i=-ncut,ncut
           ii=i+ic
           if(ii.ge.nx)ii=ii-nx
           if(ii.lt.0) ii=ii+nx
           do j=-ncut,ncut
              jj=j+jc
              if(jj.ge.ny)jj=jj-ny
              if(jj.lt.0) jj=jj+ny
              do k=-ncut,ncut
                 kk=k+kc
                 if(kk.ge.nz)kk=kk-nz
                 if(kk.lt.0) kk=kk+nz
                 ind = 1+ii+jj*nx+kk*nx*ny
                 imin = first_ngp(ind)
                 imax = imin+num_ngp(ind)-1
                 do ip = imin,imax
                    if(xsort(isort_ngp(ip)).gt.0.)then
                       is = is + 1
                       if(is > Nsel)then
                          write(*,*)'Increase Nsel in main program'
                          stop
                       end if
                       indx_sel(is) =  isort_ngp(ip)
                       x_sel(is,1) = x(isort_ngp(ip),1)
                       x_sel(is,2) = x(isort_ngp(ip),2)
                       x_sel(is,3) = x(isort_ngp(ip),3)
                       m_sel(is) = mp (isort_ngp(ip))
                       mtotsel = mtotsel + m_sel(is)
                    end if
                 end do
              end do
           end do
        end do
        npart_sel = is
!        write(*,*)'npart_sel=',npart_sel
!        write(*,*)'Mtotsel=',Mtotsel,mmin

        if(Mtotsel .ge. Mmin) then

        dt    = 1.
        niter = 0

        do while(dt.gt.dtmax.and.niter.le.100)

           dx = abs(xc-xc0)
           dx = min(dx,taillex-dx)
           dy = abs(yc-yc0)
           dy = min(dy,tailley-dy)
           dz = abs(zc-zc0)
           dz = min(dz,taillez-dz)
           dc0 = dx*dx+dy*dy+dz*dz

           ! Calcul des distances des particules au barycentre
           !--------------------------------------------------
           do i = 1,npart_sel
              xx = x_sel(i,1)
              yy = x_sel(i,2)
              zz = x_sel(i,3)
              dx = abs(xx-xc)
              dx = min(dx,taillex-dx)
              dy = abs(yy-yc)
              dy = min(dy,tailley-dy)
              dz = abs(zz-zc)
              dz = min(dz,taillez-dz)
              distance(i) = dx*dx+dy*dy+dz*dz
           end do

           ! Tri des distances
           !---------------------------------------------------
           call quick_sort(distance,isort_sel,npart_sel)

           ! Calcul du rayon de Viriel et de la masse de l'amas
           !---------------------------------------------------
           i        = 1
           overdens = 2.0*seuil
           amass    = m_sel(isort_sel(i))
           rayon    = 0.0
           do while(i.lt.10.or.(overdens.gt.seuil.and.i.lt.npart_sel.and.rayon.lt.(real(ncut)-dc0)))
              i        = i+1
              rayon    = SQRT(distance(i))
              volume   = 2./3.*twopi*rayon**3
              amass    = amass+m_sel(isort_sel(i))
              overdens = amass/volume
!              write(*,*)i,rayon,amass,overdens
           end do
           if(overdens.gt.seuil)then
              write(*,*)'Increase selection radius (-sel option in command line)'
              write(*,*)'mass     =',i
              write(*,*)'nsel     =',npart_sel
              write(*,*)'overdens =',overdens
              write(*,*)'radius   =',rayon+dc0
              write(*,*)'nsel     =',ncut
              write(*,*)'niter    =',niter
              stop
           else
              rayon =SQRT(distance(i-1))
              volume=2./3.*twopi*rayon**3
              amass =amass-m_sel(isort_sel(i))
              if(volume > 0.)then
                 overdens=amass/volume
              else
                 overdens=0.
              end if
              masse=i-1
           end if
!           write(*,*)'iter=',niter,'mass=',masse

           ! Calcul du nouveau barycentre:
           !------------------------------
           xb=0.
           yb=0.
           zb=0.
           do i=1,masse
              xx = x_sel(isort_sel(i),1)
              yy = x_sel(isort_sel(i),2)
              zz = x_sel(isort_sel(i),3)
              dx = (xx-xc)
              if(dx.gt.  taillex/2. )dx = dx-taillex
              if(dx.lt.(-taillex/2.))dx = dx+taillex
              dy = (yy-yc)
              if(dy.gt.  tailley/2. )dy = dy-tailley
              if(dy.lt.(-tailley/2.))dy = dy+tailley
              dz = (zz-zc)
              if(dz.gt.  taillez/2. )dz = dz-taillez
              if(dz.lt.(-taillez/2.))dz = dz+taillez
              xb = xb+dx*m_sel(isort_sel(i))
              yb = yb+dy*m_sel(isort_sel(i))
              zb = zb+dz*m_sel(isort_sel(i))
           enddo
           xb = xb/amass
           yb = yb/amass
           zb = zb/amass
           if(rayon > 0)then
              dt = sqrt(xb*xb+yb*yb+zb*zb)/rayon
           else
              dt = 0.
           end if
           xc = xc+xb
           yc = yc+yb
           zc = zc+zb

           niter = niter + 1
		 
           ! On sort de la boucle si la position du barycentre converge
           !-----------------------------------------------------------
        enddo

        if(amass.ge.Mmin)then
           iamas = iamas + 1
           write(*,997)iamas,masse,amass/scale**3,xc/scale,yc/scale,zc/scale &
                & ,niter,rayon/scale,overdens
        end if

        ! On retire de la liste des particules celles de l'amas
        !------------------------------------------------------
        do i = 1,masse
!           structure(indx_sel(isort_sel(i))) = iamas
           xsort(indx_sel(isort_sel(i))) = 0.
        enddo
	      
        ! Ecriture des caracteristiques de l'amas sur fichier
        !----------------------------------------------------
        if(amass.ge.Mmin)then
           write(10,999)iamas,masse,amass/scale**3,xc/scale,yc/scale,zc/scale
        end if

     end if

     end if
	   
     if(mod(ipart,100000)==0)write(*,'(0PF5.1,"% complete")')100.*dble(ipart)/dble(npart_new)

  end do
  
  close(10)

!  nomf = 'st_sod_'//TRIM(repository)//'.dat'
!  open (10,file=nomf,form='unformatted')
!  write(10)structure
!  close(10)

997 format (I8,1x,I8,4(1x,1pe10.3),1x,I2,1x,5(E10.3,1x))
999 format (I8,1x,I8,4(1x,1pe10.3))
	
  return
end subroutine spherover

subroutine title(n,nchar)
  integer::n
  character*5::nchar

  character*1::nchar1
  character*2::nchar2
  character*3::nchar3
  character*4::nchar4
  character*5::nchar5


  if(n.ge.10000)then
     write(nchar5,'(i5)') n
     nchar = nchar5
  elseif(n.ge.1000)then
     write(nchar4,'(i4)') n
     nchar = '0'//nchar4
  elseif(n.ge.100)then
     write(nchar3,'(i3)') n
     nchar = '00'//nchar3
  elseif(n.ge.10)then
     write(nchar2,'(i2)') n
     nchar = '000'//nchar2
  else
     write(nchar1,'(i1)') n
     nchar = '0000'//nchar1
  endif
end subroutine title

SUBROUTINE quick_sort(list, order, n)
  
  ! Quick sort routine from:
  ! Brainerd, W.S., Goldberg, C.H. & Adams, J.C. (1990) "Programmer's Guide to
  ! Fortran 90", McGraw-Hill  ISBN 0-07-000248-7, pages 149-150.
  ! Modified by Alan Miller to include an associated integer array which gives
  ! the positions of the elements in the original order.
  
  IMPLICIT NONE
  INTEGER :: n
  REAL*8, DIMENSION (1:n), INTENT(INOUT)  :: list
  INTEGER, DIMENSION (1:n), INTENT(OUT)  :: order
  
  ! Local variable
  INTEGER :: i
  
  DO i = 1, n
     order(i) = i
  END DO
  
  CALL quick_sort_1(1, n)
  
CONTAINS
  
  RECURSIVE SUBROUTINE quick_sort_1(left_end, right_end)
    
    INTEGER, INTENT(IN) :: left_end, right_end
    
    !     Local variables
    INTEGER             :: i, j, itemp
    REAL*8              :: reference, temp
    INTEGER, PARAMETER  :: max_simple_sort_size = 6
    
    IF (right_end < left_end + max_simple_sort_size) THEN
       ! Use interchange sort for small lists
       CALL interchange_sort(left_end, right_end)
       
    ELSE
       ! Use partition ("quick") sort
       reference = list((left_end + right_end)/2)
       i = left_end - 1; j = right_end + 1
       
       DO
          ! Scan list from left end until element >= reference is found
          DO
             i = i + 1
             IF (list(i) >= reference) EXIT
          END DO
          ! Scan list from right end until element <= reference is found
          DO
             j = j - 1
             IF (list(j) <= reference) EXIT
          END DO
          
          
          IF (i < j) THEN
             ! Swap two out-of-order elements
             temp = list(i); list(i) = list(j); list(j) = temp
             itemp = order(i); order(i) = order(j); order(j) = itemp
          ELSE IF (i == j) THEN
             i = i + 1
             EXIT
          ELSE
             EXIT
          END IF
       END DO
       
       IF (left_end < j) CALL quick_sort_1(left_end, j)
       IF (i < right_end) CALL quick_sort_1(i, right_end)
    END IF
    
  END SUBROUTINE quick_sort_1
  
  
  SUBROUTINE interchange_sort(left_end, right_end)
    
    INTEGER, INTENT(IN) :: left_end, right_end
    
    !     Local variables
    INTEGER             :: i, j, itemp
    REAL*8              :: temp
    
    DO i = left_end, right_end - 1
       DO j = i+1, right_end
          IF (list(i) > list(j)) THEN
             temp = list(i); list(i) = list(j); list(j) = temp
             itemp = order(i); order(i) = order(j); order(j) = itemp
          END IF
       END DO
    END DO
    
  END SUBROUTINE interchange_sort
  
END SUBROUTINE quick_sort
